import os
from os import path
from pathlib import Path
import re
import logging


marker_regex = re.compile(r'(.*)marker(.*).txt$', flags=re.I)

local_config_file = Path(__file__).resolve().with_name("config.cfg")
dragndrop_bat_file = Path(__file__).resolve().parent / "scripts" / "dragndrop.bat"


def is_marker_file(arg):
    return marker_regex.match(path.basename(arg))


def is_mosaic_file(arg):
    return os.path.splitext(arg)[1].lower() == ".txt" and get_mrc_file(arg)


def argument_organiser(arguments):
    logging.debug("Testing args: %s", arguments)
    args_out = [None, ]
    for arg in arguments:
        if is_marker_file(arg):
            args_out.append(arg)
        elif is_mosaic_file(arg):
            args_out[0] = arg
        else:
            logging.warning("argument %s is invalid", arg)
    logging.debug("Returning args: %s", args_out)
    return args_out


def get_mrc_file(arg, return_data=False):
    logging.debug("Opening file: %s", arg)
    txt_path = Path(arg)
    if txt_path.is_file():
        with txt_path.open('rb') as csvfile:
            mrc_path = csvfile.readline().rstrip().decode('utf-8')
            if return_data:
                from numpy import genfromtxt
                location_array = genfromtxt(csvfile, delimiter=",")
        
        if path.splitext(mrc_path)[1].lower() == ".mrc":
            if not path.exists(mrc_path):
                msg = (
                    "Cannot find %s, so the current directory will be tried",
                    mrc_path)
                if return_data:
                    logging.warning(*msg)
                else:
                    logging.debug(*msg)
                if "\\" in mrc_path:
                    #  Windows file paths
                    mrc_name  = mrc_path.split("\\")[-1]
                elif "/" in mrc_path:
                    # Unix file paths
                    mrc_name = mrc_path.split("/")[-1]
                else:
                    mrc_name =  path.basename(mrc_path)
                mrc_path = txt_path.parent / mrc_name
                if not mrc_path.exists():
                    if return_data:
                        raise IOError(f"Cannot find path {mrc_path}")
                    logging.error("%s cannot be found", mrc_path)
                    return False
            
            if return_data:
                return path.abspath(mrc_path), location_array
            return True
        
        elif return_data:
            raise IOError(f"Expected an mrc file, instead the txt file linked to {mrc_path}")
        logging.error("Mosaic txt file contains invalid mrc path: %s", mrc_path)
    
    elif return_data:
        raise IOError(f"Mosaic txt file {txt_path} does not exist")
    logging.error("Invalid txt file: %s", arg)
    return False


def _get_Windows_home_path():
    home = os.getenv("USERPROFILE")
    if home is None:
        homedrive = os.getenv("HOMEDRIVE")
        homepath = os.getenv("HOMEPATH")
        if homedrive and homepath:
            home = os.path.join(homedrive, homepath)
        else:
            home = os.getenv("HOME")

    if home is not None and "Users" in home:
        try:
            home_path = Path(home).resolve(strict=True)
            return home_path
        except FileNotFoundError:
            pass
    logging.error("Cannot find valid home path. The following path was found: '%s'", home)
    raise FileNotFoundError(f"Cannot find valid home path. The following path was found: '{home}''")


def get_user_log_path():
    try:
        if os.name == "nt":
            user_log_path = _get_Windows_home_path() / "AppData" / "Local" / "Temp" / "stitchm"
            # Windows options are read-only or read and write (default), so no mode set
            user_log_path.mkdir(exist_ok=True)
        elif os.name == "posix":
            from tempfile import mkdtemp
            from getpass import getuser
            # Makes user only access dir:
            user_log_path = Path(mkdtemp(prefix=f"stitchm_log_{getuser()}", dir="/tmp"))
            # User can read/write, group can read, others can't access:
            user_log_path.chmod(mode=0o640)
        else:
            logging.error("Operating system cannot be determined")
        
        logging.info("Creating log files in path %s.", str(user_log_path))
        return user_log_path
    except Exception:
        logging.error(f"Error occurred creating log files", exc_info=True)
        return None
    

def get_user_config_path():
    logging_messages = []
    if os.name == "nt":
        user_config_path = _get_Windows_home_path() / "AppData" / "Local" / "stitchm" / "config.cfg"
    elif os.name == "posix":
        user_config_path = Path.home() / ".config" / "stitchm" / "config.cfg"
    else:
        logging_messages.append("Operating system cannot be determined")
        return None, logging_messages
    return user_config_path, logging_messages


def get_config():
    from configparser import ConfigParser
    config_messages = []
    config = ConfigParser()
    user_config_file, logging_messages = get_user_config_path()
    config_messages.extend(logging_messages)
    if user_config_file is not None and user_config_file.exists():
        try:
            with open(user_config_file) as f:
                config.read_file(f)
            return config, config_messages
        except Exception:
            config_messages.append(f"Opening user config file failed. Please delete your existing file and try again! (Expected path: {user_config_file})")
    with open(local_config_file) as f:
            config.read_file(f)
    return config, config_messages


def create_user_config():
    from shutil import copyfile
    user_config_file, logging_messages = get_user_config_path()
    for message in logging_messages:
        logging.warning(message)
    if user_config_file is not None:
        try:
            Path.mkdir(user_config_file.parent, parents=False, exist_ok=True)
            logging.info("Creating user config file in path %s.", str(user_config_file))
            copyfile(local_config_file, user_config_file)
            logging.info("User config file has been created in %s. This file will override default settings.", str(user_config_file))
        except Exception:
            logging.error("Unable to create user config file due to directory issues", exc_info=True)
    else:
        logging.error("Unable to create user config file")


def boolean_config_handler(config, section, key, default):
    try:
        return config.getboolean(section, key, fallback=default)
    except ValueError:
        logging.error("Invalid '%s' option found, default of '%s' will be used", key, default)
        return default


def _create_lnk_file(shortcut_path):
    try:
        # win23com is from the package pywin32, only available in Windows
        import win32com.client
    except ImportError:
        msg = "win32com of pywin32 cannot be imported! Please run 'pip install pywin32' (with '--user' argument if on a shared python environment) then try again."
        print(msg)
        logging.error(msg)
        raise
    shell = win32com.client.Dispatch("WScript.Shell")
    shortcut = shell.CreateShortCut(str(shortcut_path))
    shortcut.Targetpath = str(dragndrop_bat_file)
    shortcut.save()
    msg = f"Shortcut created! It can be found here: {shortcut_path}"
    print(msg)
    logging.info(msg)


def create_Windows_shortcut():
    if os.name != "nt":
        logging.error("This command is only valid on Windows installations.")
        return 
    else:
        # Place link on users desktop
        shortcut_path = _get_Windows_home_path() / "Desktop" / "StitchM.lnk"
        msg = f"Creating shortcut on user desktop: {shortcut_path}"
        logging.info(msg)
        if shortcut_path.exists():
            msg = f"StitchM shortcut found:'{shortcut_path}'. Are you sure you want to replace it? (y/N)"
            user_input = str(input(msg))
            logging.debug(msg)
            logging.debug("User input: %s", user_input)
            if user_input.lower() == "y" or user_input.lower() == "yes":
                logging.info("The existing shortcut will be replaced.")
            elif user_input.lower() == "n" or user_input.lower() == "no":
                logging.info("The existing shortcut will not be modified.")
            else:
                logging.info("Invalid input: %s. The existing shortcut will not be modified.", user_input)
                return
        _create_lnk_file(shortcut_path)



import sys
import logging

from .__init__ import __version__


def sort_args_then_run():
    from .log_handler import LogHandler
    from .file_handler import argument_organiser, get_config, create_user_config, create_Windows_shortcut, boolean_config_handler
    try:
        argv = sys.argv[1:]
        # Allows setup options without StitchM script interface:
        create_usr_config = "-cfg" in argv or "--config" in argv
        create_win_shortcut = "-win" in argv or "--windows-shortcut" in argv
        if create_usr_config or create_win_shortcut:
            with LogHandler("info", "info"):
                if create_usr_config:
                    create_Windows_shortcut()
                if create_win_shortcut:
                    create_user_config()

        config, config_messages = get_config()
        with LogHandler(config=config, config_messages=config_messages):
            logging.info("StitchM given arguments: %s", argv)
            args = argument_organiser(argv)
            logging.info("Sorted arguments: %s", args)
            if args[0] is not None:
                if len(args) > 2:
                    args = args[0:2]
                main_run(config, *args)
                if boolean_config_handler(config, 'OTHER', 'wait upon completion', default=False):
                    input("Processing complete! Press enter to exit")
                return
            else:
                logging.error("No valid mosaic file")
    except IOError:
        logging.error('Error has occurred while stitching or saving mosaic. Please see traceback for more info.', exc_info=True)
    except Exception:
        logging.error("Unknown error occurred. Please see traceback for more info.", exc_info=True)
    if boolean_config_handler(config, 'OTHER', 'wait upon failure', default=True):
        input("Processing failed! Press enter to exit")

def _stitch(config, mosaic, markers, normalise):
    from pathlib import Path
    from .file_handler import is_mosaic_file, is_marker_file, boolean_config_handler
    from .unstitched_image import UnstitchedImage
    from .metadata_maker import MetadataMaker
    from .stitcher import Stitcher
    dtype = "uint16"
    try:
        if is_mosaic_file(mosaic):
            mosaic_path = Path(mosaic).resolve()  # Gets absolute path of mosaic file
            tiff_path = mosaic_path.with_suffix(".ome.tiff")

            unstitched = UnstitchedImage(mosaic_path)
            stitcher = Stitcher(dtype)
            mosaic_array = stitcher.make_mosaic(unstitched, boolean_config_handler(config, 'PROCESSING', 'filter', default=True), normalise)
            metadata_creator = MetadataMaker(tiff_path.name, unstitched, dtype)

            if markers is not None and is_marker_file(markers) and Path(markers).is_file():
                tiff_name = tiff_path.name.replace(".ome.tiff", "_marked.ome.tiff")
                tiff_path = tiff_path.parent / tiff_name
                metadata_creator.add_markers(tiff_name, markers)
            return mosaic_array, metadata_creator.get(), str(tiff_path)
        else:
            logging.error("Mosaic file path cannot be resolved")
            raise IOError("Mosaic file path cannot be resolved")
    except Exception:
        arg_string = ", ".join((str(mosaic or None), str(markers or None), str(normalise)))
        logging.error("Invalid arguments: %s", arg_string, exc_info=True)
        if boolean_config_handler(config, 'OTHER', 'wait upon failure', default=True):
            input("Processing failed! Press enter to exit")
        raise IOError("Invalid arguments: %s", arg_string)
    

def _save(mosaic, metadata, tiff_filename):
    import tifffile as tf
    from numpy import iinfo, uint32
    try:
        logging.info("Saving %s", tiff_filename)
        bigtiff = mosaic.size * mosaic.itemsize >= iinfo(uint32).max  # Check if data bigger than 4GB TIFF limit
        with tf.TiffWriter(tiff_filename, bigtiff=bigtiff, ome=False) as tif:
            tif.write(
                mosaic,
                description=metadata.to_xml().encode(),
                photometric='MINISBLACK',
                metadata={'axes':'YX'},
                software=f"StitchM {__version__}")
    except Exception:
        logging.error("Cannot save: %s", tiff_filename, exc_info=True)
        raise IOError("Cannot save: {}".format(tiff_filename))

def main_run(config, mosaic, markers=None, normalise=True):
    """
    PARAMETERS:
        mosaic - Path of .txt file that contains the mosaic information, including the path to the .mrc file
        markers - Path of .txt file that contains a list of marker placements and associated numbers (please make sure this correctly corresponds to the mosaic file)
        normaliseOff - boolean that specifies if we should not normalise the image. 
    
    The output will be saved as the mosaic filename, with the suffix '.ome.tiff' (or '_marked.ome.tiff' if markers are supplied), in same directory as the mosaic file.
    """
    logging.info("Running StitchM with arguments: mosaic=%s, markers=%s, normalise=%s", mosaic, markers, normalise)
    mosaic, metadata, tiff_file = _stitch(config, mosaic, markers, normalise)
    _save(mosaic, metadata, tiff_file)


import unittest
from unittest.mock import patch, mock_open
import os
from pathlib import Path
from numpy import genfromtxt
from numpy.testing import assert_array_equal

from stitch_m import file_handler

base_path = Path("/dls/science/groups/das/ExampleData/B24_test_data/StitchM_test_data/files/")
test_file = base_path / "Fid_T2G3_mosaic.txt"
bad_path = Path("/bad/path/to/no-files/")
mrc_filename = "Fid_T2G3_mosaic.mrc"
test_marker_file = base_path / "Fid_T2G3_markers.txt"
bad_file = base_path / "bad_file.txt"

testargs = [test_file, test_marker_file]
bad_testargs = [bad_file, test_marker_file, test_file]

test_config = Path(__file__).resolve().with_name("config.cfg")


@unittest.skipUnless(base_path.exists(), "base path cannot be accessed")
class FileHandlerTests(unittest.TestCase):

    @patch('stitch_m.file_handler')
    def test_argument_filtering(self, mocked_file_handler):
        mocked_file_handler.local_config_file.return_value = test_config
        if os.path.exists(base_path):
            for arguments in [testargs, testargs[::-1], bad_testargs]:
                arguments_out = file_handler.argument_organiser(arguments)
                self.assertEqual(arguments_out, testargs)
        else:
            print("Cannot run test without access to dls directories")

    @patch("logging.warning")
    @patch("logging.info")
    def test_is_mosaic_file(self, mocked_logging_info, mocked_logging_warning):
        self.assertTrue(
            file_handler.is_mosaic_file(test_file),
            msg=f"{test_file.name} {'exists' if test_file.exists() else 'does not exist'}")
        self.assertFalse(
            file_handler.is_mosaic_file(bad_file),
            msg=f"{bad_file.name} {'exists' if bad_file.exists() else 'does not exist'}")
        mocked_logging_info.assert_not_called()
        mocked_logging_warning.assert_not_called()

    @patch("pathlib.Path.open", new_callable=mock_open, read_data=(str(bad_path / mrc_filename) + "\n").encode())
    def test_get_mrc_file_opens_file_bad_path(self, mock_open):
        with test_file.open('rb') as csvfile:
            mrc_path = Path(csvfile.readline().rstrip().decode('utf-8')).absolute()
            expected_location_array = genfromtxt(csvfile, delimiter=",")

        expected_output_path = str(base_path / mrc_path.name)
        
        output_path, location_array = file_handler.get_mrc_file(test_file, True)

        self.assertEqual(output_path, expected_output_path)
        assert_array_equal(location_array, expected_location_array)

    @patch("pathlib.Path.open", new_callable=mock_open, read_data=(str(base_path / mrc_filename) + "\n").encode())
    def test_get_mrc_file_opens_file_good_path(self, mock_open):
        with test_file.open('rb') as csvfile:
            mrc_path = Path(csvfile.readline().rstrip().decode('utf-8')).absolute()
            expected_location_array = genfromtxt(csvfile, delimiter=",")

        expected_output_path = str(base_path / mrc_path.name)
        
        output_path, location_array = file_handler.get_mrc_file(test_file, True)

        self.assertEqual(output_path, expected_output_path)
        assert_array_equal(location_array, expected_location_array)

    @patch("pathlib.Path.open", new_callable=mock_open, read_data=(str(bad_path / mrc_filename) + "\n").encode())
    def test_get_mrc_file_good_file_with_bad_path(self, mock_open):
        with test_file.open('rb') as csvfile:
            mrc_path = Path(csvfile.readline().rstrip().decode('utf-8')).absolute()
        
        self.assertTrue(file_handler.get_mrc_file(test_file, False))

    @patch("pathlib.Path.open", new_callable=mock_open, read_data=(str(base_path / mrc_filename) + "\n").encode())
    def test_get_mrc_file_good_file_with_good_path(self, mock_open):
        with test_file.open('rb') as csvfile:
            mrc_path = Path(csvfile.readline().rstrip().decode('utf-8')).absolute()

        self.assertTrue(file_handler.get_mrc_file(test_file, False))


    def test_get_mrc_file_bad_file(self):
        self.assertFalse(file_handler.get_mrc_file(bad_file, False))



import unittest
from unittest.mock import patch
from numpy.testing import assert_array_equal

import os
import tifffile as tf
import numpy as np
from pathlib import Path


from stitch_m import file_handler, stitch_and_save

# To quickly update the files:
# - Comment out file removal in tearDownClass
# - Comment out file opening and comparisons
# - Delete old expected_output files
# - Within the base_path dir, run:
#   for f in *.ome.tiff; do mv "$f" `echo ${f} | sed 's/.ome.tiff/_expected_output.ome.tiff/'`; done
# - Uncommented out code commented out above

base_path = Path("/dls/science/groups/das/ExampleData/B24_test_data/StitchM_test_data/files")
test_files = (
    base_path / "B15Grid2.txt",
    base_path / "B8G1-IR_mosaic.txt",
    base_path / "B8G2-IR_mosaic.txt",
    base_path / "Fid_T2G3_mosaic.txt",
    base_path / "Yo10_G3_mosaic.txt"
    )

test_marker_files = (
    base_path / "B15_location_markers.txt",
    base_path / "B8G1-IR_markers.txt",
    base_path / "B8G2-IR_markers.txt",
    base_path / "Fid_T2G3_markers.txt",
    base_path / "Yo10_G3_mosaic_MARKERS.txt"
    )

# # Reduced tuple for quick tests:
# test_files = (
#     base_path / "Fid_T2G3_mosaic.txt",
#     )
# test_marker_files = (
#     base_path / "Fid_T2G3_markers.txt",
#     )
test_files = [str(path) for path in test_files]
test_marker_files = [str(path) for path in test_marker_files]

expected_outputs = [
    path.replace(".txt", "_expected_output.ome.tiff")
    for path in test_files]
expected_marked_outputs = [
    path.replace(".txt", "_marked_expected_output.ome.tiff")
    for path in test_files]
test_config = Path(__file__).resolve().with_name("config.cfg")


@unittest.skipUnless(base_path.exists(), "base path cannot be accessed")
class EndToEndTests(unittest.TestCase):

    @classmethod
    def tearDownClass(cls):
        for f in test_files:
            output_path = f.replace('.txt', '.ome.tiff')
            if os.path.isfile(output_path):
                os.remove(output_path)
            output_path = f.replace('.txt', '_marked.ome.tiff')
            if os.path.isfile(output_path):
                os.remove(output_path)

    @patch('stitch_m.file_handler')
    def test_end_to_end_simple(self, mocked_file_handler):
        mocked_file_handler.local_config_file.return_value = test_config
        for i in range(len(test_files)):
            test_file = test_files[i]
            stitch_and_save(test_file)
            output_path = test_file.replace('.txt', '.ome.tiff')
            self.assertTrue(os.path.isfile(output_path), msg=f"{output_path} not found")
            output_image = tf.imread(output_path)

            expected_file = expected_outputs[i]
            expected_image = tf.imread(expected_file)

            assert_array_equal(output_image, expected_image)

    @patch('stitch_m.file_handler')
    def test_end2end_with_markers(self, mocked_file_handler):
        self.maxDiff = 8000  # Set diff so that all metadata can be read
        mocked_file_handler.local_config_file.return_value = test_config
        for i in range(len(test_files)):
            test_file = test_files[i]
            stitch_and_save(test_file, test_marker_files[i])
            output_path = test_file.replace('.txt', '_marked.ome.tiff')
            self.assertTrue(os.path.isfile(output_path), msg=f"{output_path} not found")
            
            with tf.TiffFile(output_path) as tif:
                output_image = tif.asarray()
                output_metadata = tif.ome_metadata

            with tf.TiffFile(expected_marked_outputs[i]) as tif:
                expected_image = tif.asarray()
                expected_metadata = tif.ome_metadata

            assert_array_equal(output_image, expected_image)
            self.assertEqual(
                output_metadata, expected_metadata,
                msg=f"Metadata should match for {output_path} & {expected_marked_outputs[i]}\n")
